# -*- coding: utf-8 -*-

cpdef str get_rule_month(object source, str default=*)

cpdef get_freq_code(freqstr)
cpdef object get_freq(object freq)
cpdef str get_base_alias(freqstr)
cpdef int get_to_timestamp_base(int base)
cpdef str get_freq_str(base, mult=*)
